# **************************************************************************** #
#                                                                              #
#                                                         :::      ::::::::    #
#    boundary_terms.py                                  :+:      :+:    :+:    #
#                                                     +:+ +:+         +:+      #
#    By: daniloceano <danilo.oceano@gmail.com>      +#+  +:+       +#+         #
#                                                 +#+#+#+#+#+   +#+            #
#    Created: 2022/01/31 20:15:59 by daniloceano       #+#    #+#              #
#    Updated: 2024/07/18 00:21:16 by daniloceano      ###   ########.fr        #
#                                                                              #
# **************************************************************************** #

"""
This script defines the CBoundaryTerms object. It uses the MetData object as
an input. The built-in functions use the input data to compute the following
boundary terms of the Lorenz Energy Cycle.

Created by:
    Danilo Couto de Souza
    Universidade de São Paulo (USP)
    Instituto de Astornomia, Ciências Atmosféricas e Geociências
    São Paulo - Brazil

Contact:
    danilo.oceano@gmail.com
"""

import logging

import numpy as np
from metpy.constants import Re, g

from ..utils.box_data import BoxData
from ..utils.calc_averages import CalcAreaAverage, CalcZonalAverage


class BoundaryTerms:
    """
    Class to compute boundary terms of the Lorenz Energy Cycle.

    Attributes:
        method (str): The computation method used ('fixed', 'track', or 'choose').
        box_obj (BoxData): The BoxData object containing meteorological data.

    Methods:
        calc_baz(self): flux of zonal available potential energy across the boundaries (BAZ).
        calc_bae(self): flux of eddy available potential energy across the boundaries (BAE).
        calc_bkz(self): flux of zonal kinetic energy across the boundaries (BKZ).
        calc_bke(self): flux of eddy kinetic energy across the boundaries (BKE).
        calc_boz(self): appearence of zonal kinetic energywith work produced at the boundaries (BΦZ).
        calc_boe(self): appearence of eddy kinetic energy with work produced at the boundaries (BΦE).

    Sources for formulas used here:
        Michaelides, S. C. (1987).
        Limited Area Energetics of Genoa Cyclogenesis,
        Monthly Weather Review, 115(1), 13-26. From:
        https://journals.ametsoc.org/view/journals/mwre/115/1/1520-0493_1987_115_0013_laeogc_2_0_co_2.xml

        Brennan, F. E., & Vincent, D. G. (1980).
        Zonal and Eddy Components of the Synoptic-Scale Energy Budget
        during Intensification of Hurricane Carmen (1974),
        Monthly Weather Review, 108(7), 954-965. Retrieved Jan 25, 2022, from:
        https://journals.ametsoc.org/view/journals/mwre/108/7/1520-0493_1980_108_0954_zaecot_2_0_co_2.xml
    """

    def __init__(self, box_obj: BoxData, method: str, app_logger: logging.Logger):
        """Initialize the EnergyContents object with a BoxData object and a method."""
        self._initialize_attributes(box_obj, method, app_logger)

    def _initialize_attributes(self, box_obj, method, app_logger):
        """Helper method to initialize attributes from the BoxData object."""
        # Operational attributes
        self.method = method
        self.box_obj = box_obj
        self.results_subdirectory = box_obj.results_subdirectory
        self.app_logger = app_logger

        # Initialize spatial and temporal attributes
        self.LonIndexer = box_obj.LonIndexer
        self.LatIndexer = box_obj.LatIndexer
        self.western_limit = box_obj.western_limit
        self.eastern_limit = box_obj.eastern_limit
        self.southern_limit = box_obj.southern_limit
        self.northern_limit = box_obj.northern_limit
        self.VerticalCoordIndexer = box_obj.VerticalCoordIndexer
        self.TimeName = box_obj.TimeName

        # Initialize lengths for averaging
        self.xlength = box_obj.xlength
        self.ylength = box_obj.ylength

        # Initialize pressure data
        self.PressureData = box_obj.PressureData

        # Initialize attributes related to temperature
        self.tair_AE = box_obj.tair_AE
        self.tair_ZE = box_obj.tair_ZE

        # Initialize attributes related to wind components
        self.u = box_obj.u
        self.u_ZA = box_obj.u_ZA
        self.u_ZE = box_obj.u_ZE
        self.v = box_obj.v
        self.v_ZA = box_obj.v_ZA
        self.v_ZE = box_obj.v_ZE

        # Initialize attributes related to vertical velocity
        self.omega = box_obj.omega
        self.omega_ZE = box_obj.omega_ZE
        self.omega_AA = box_obj.omega_AA
        self.omega_ZA = box_obj.omega_ZA
        self.omega_AE = box_obj.omega_AE

        # Initialize static stability parameter
        self.sigma_AA = box_obj.sigma_AA

        # Initialize attributes related to geopotential
        self.geopt_ZE = box_obj.geopt_ZE
        self.geopt_AE = box_obj.geopt_AE

        # Initialize operand using the notation from Michaelides (1987)
        self.c1 = -1 / (Re * self.xlength * self.ylength)
        self.c2 = -1 / (Re * self.ylength)

    def calc_baz(self):
        """
        Computes the flux of Zonal Available Potential Energy across the boundaries (BAZ).

        Note:
            On Brennan et al. (1980), the area average of term3 is missing (Muench, 1965).
        """
        self.app_logger.debug("Calculating BAZ...")

        # First term
        term1 = (
            (2 * self.tair_AE * self.tair_ZE * self.u) + (self.tair_AE**2 * self.u)
        ) / (2 * self.sigma_AA)
        term1 = term1.sel(**{self.LonIndexer: self.eastern_limit}) - term1.sel(
            **{self.LonIndexer: self.western_limit}
        )
        term1 = term1.integrate("rlats")
        term1 = self._handle_nans(term1)
        term1 = (
            term1.integrate(self.VerticalCoordIndexer)
            * self.PressureData.metpy.units
            * self.c1
        )

        # Second term
        term2 = self.v_ZE * self.tair_ZE
        term2 = CalcZonalAverage(term2, self.xlength) * 2 * self.tair_AE
        term2 = (term2 + ((self.tair_AE**2) * self.v_ZA)) * self.tair_AE["coslats"]
        term2 = (
            term2.sel(**{self.LatIndexer: self.northern_limit})
            - term2.sel(**{self.LatIndexer: self.southern_limit})
        ) / (2 * self.sigma_AA)
        term2 = self._handle_nans(term2)
        term2 = (
            term2.integrate(self.VerticalCoordIndexer)
            * self.PressureData.metpy.units
            * self.c2
        )

        # Third term
        term3a = 2 * self.omega_ZE * self.tair_ZE
        term3a = CalcZonalAverage(term3a, self.xlength) * self.tair_AE
        term3b = self.omega_ZA * self.tair_AE**2
        term3 = term3a + term3b
        term3 = self._handle_nans(term3)
        term3 = CalcAreaAverage(term3, self.ylength) / (2 * self.sigma_AA)
        term3 = term3.isel(**{self.VerticalCoordIndexer: -1}) - term3.isel(
            **{self.VerticalCoordIndexer: 0}
        )

        # Combine terms and save the result
        function = term1 + term2 - term3
        function = self._handle_nans(function)
        Baz = self._convert_units(function, "BAZ")

        self.app_logger.debug("Done.")
        return Baz

    def calc_bae(self):
        """Computes the flux of Eddy Available Potential Energy across the boundaries (BAE)."""
        self.app_logger.debug("Calculating BAE...")

        # First Integral
        term1 = self.u * (self.tair_ZE**2)
        term1 = term1.sel(**{self.LonIndexer: self.eastern_limit}) - term1.sel(
            **{self.LonIndexer: self.western_limit}
        )
        term1 = (term1 / (2 * self.sigma_AA)).integrate("rlats")
        term1 = self._handle_nans(term1)
        term1 = (
            term1.integrate(self.VerticalCoordIndexer)
            * self.PressureData.metpy.units
            * self.c1
        )

        # Second Integral
        term2 = (
            CalcZonalAverage(self.v * self.tair_ZE**2, self.xlength)
            * self.tair_ZE["coslats"]
        )
        term2 = term2 / (2 * self.sigma_AA)
        term2 = term2.sel(**{self.LatIndexer: self.northern_limit}) - term2.sel(
            **{self.LatIndexer: self.southern_limit}
        )
        term2 = self._handle_nans(term2)
        term2 = (
            term2.integrate(self.VerticalCoordIndexer)
            * self.PressureData.metpy.units
            * self.c2
        )

        # Third Term
        term3 = (self.omega * self.tair_ZE**2) / (2 * self.sigma_AA)
        term3 = CalcAreaAverage(term3, self.ylength, xlength=self.xlength)
        term3 = self._handle_nans(term3)
        term3 = term3.isel(**{self.VerticalCoordIndexer: -1}) - term3.isel(
            **{self.VerticalCoordIndexer: 0}
        )

        # Combine terms and save the result
        function = term1 + term2 - term3
        function = self._handle_nans(function)
        Bae = self._convert_units(function, "BAE")

        self.app_logger.debug("Done.")
        return Bae

    def calc_bkz(self):
        """Computes the flux of Zonal Kinetic Energy across the boundaries (BKZ)."""
        self.app_logger.debug("Calculating BKZ...")

        # First term
        term1 = self.u * (self.u**2 + self.v**2 - self.u_ZE**2 - self.v_ZE**2)
        term1 = term1.sel(**{self.LonIndexer: self.eastern_limit}) - term1.sel(
            **{self.LonIndexer: self.western_limit}
        )
        term1 = (term1 / (2 * g)).integrate("rlats")
        term1 = self._handle_nans(term1)
        term1 = (
            term1.integrate(self.VerticalCoordIndexer)
            * self.PressureData.metpy.units
            * self.c1
        )

        # Second term
        term2 = (
            (self.u**2 + self.v**2 - self.u_ZE**2 - self.v_ZE**2)
            * self.v
            * self.v["coslats"]
        )
        term2 = CalcZonalAverage(term2, self.xlength)
        term2 = term2.sel(**{self.LatIndexer: self.northern_limit}) - term2.sel(
            **{self.LatIndexer: self.southern_limit}
        )
        term2 = self._handle_nans(term2)
        term2 = (
            (term2 / (2 * g)).integrate(self.VerticalCoordIndexer)
            * self.PressureData.metpy.units
            * self.c2
        )

        # Third term
        term3 = (self.u**2 + self.v**2 - self.u_ZE**2 - self.v_ZE**2) * self.omega
        term3 = CalcAreaAverage(term3, self.ylength, xlength=self.xlength) / (2 * g)
        term3 = self._handle_nans(term3)
        term3 = term3.isel(**{self.VerticalCoordIndexer: -1}) - term3.isel(
            **{self.VerticalCoordIndexer: 0}
        )

        # Combine terms and save the result
        function = term1 + term2 - term3
        function = self._handle_nans(function)
        Bkz = self._convert_units(function, "BKz")

        self.app_logger.debug("Done.")
        return Bkz

    def calc_bke(self):
        """Computes the flux of Eddy Kinetic Energy across the boundaries (BKE)."""
        self.app_logger.debug("Calculating BKE...")

        # First term
        term1 = self.u * (self.u_ZE**2 + self.v_ZE**2)
        term1 = term1.sel(**{self.LonIndexer: self.eastern_limit}) - term1.sel(
            **{self.LonIndexer: self.western_limit}
        )
        term1 = (term1 / (2 * g)).integrate("rlats")
        term1 = self._handle_nans(term1)
        term1 = (
            term1.integrate(self.VerticalCoordIndexer)
            * self.PressureData.metpy.units
            * self.c1
        )

        # Second term
        term2 = (self.u_ZE**2 + self.v_ZE**2) * self.v * self.v["coslats"]
        term2 = CalcZonalAverage(term2, self.xlength)
        term2 = term2.sel(**{self.LatIndexer: self.northern_limit}) - term2.sel(
            **{self.LatIndexer: self.southern_limit}
        )
        term2 = self._handle_nans(term2)
        term2 = (
            (term2 / (2 * g)).integrate(self.VerticalCoordIndexer)
            * self.PressureData.metpy.units
            * self.c2
        )

        # Third term
        term3 = (self.u_ZE**2 + self.v_ZE**2) * self.omega
        term3 = CalcAreaAverage(term3, self.ylength, xlength=self.xlength) / (2 * g)
        term3 = self._handle_nans(term3)
        term3 = term3.isel(**{self.VerticalCoordIndexer: -1}) - term3.isel(
            **{self.VerticalCoordIndexer: 0}
        )

        # Combine terms and save the result
        function = term1 + term2 - term3
        function = self._handle_nans(function)
        Bke = self._convert_units(function, "BKE")

        self.app_logger.debug("Done.")
        return Bke

    def calc_boz(self):
        """
        Computes the appearence of Zonal Kinetic Energy associated with work produced at its boundaries (BΦZ).

        Note: Cannot perform eastern boundary minus western boundary on the first term
        """
        self.app_logger.debug("Calculating BΦZ...")

        # First term
        term1 = (self.v_ZA * self.geopt_AE) / g
        term1 = term1.integrate("rlats")
        term1 = self._handle_nans(term1)
        term1 = (
            term1.integrate(self.VerticalCoordIndexer)
            * self.PressureData.metpy.units
            * self.c1
        )

        # Second term
        term2 = (self.v_ZA * self.geopt_AE) * self.v_ZA["coslats"] / g
        term2 = term2.sel(**{self.LatIndexer: self.northern_limit}) - term2.sel(
            **{self.LatIndexer: self.southern_limit}
        )
        term2 = self._handle_nans(term2)
        term2 = (
            term2.integrate(self.VerticalCoordIndexer)
            * self.PressureData.metpy.units
            * self.c2
        )

        # Third term
        term3 = CalcAreaAverage(self.omega_AE * self.geopt_AE, self.ylength) / g
        term3 = self._handle_nans(term3)
        term3 = term3.isel(**{self.VerticalCoordIndexer: -1}) - term3.isel(
            **{self.VerticalCoordIndexer: 0}
        )

        function = term1 + term2 - term3
        function = self._handle_nans(function)
        Boz = self._convert_units(function, "BOZ")

        self.app_logger.debug("Done.")
        return Boz

    def calc_boe(self):
        """Computes the appearence of Eddy Kinetic Energy associated with work produced at its boundaries (BΦE)."""
        self.app_logger.debug("Calculating BΦE...")

        # First term
        term1 = (self.v_ZE * self.geopt_AE) / g
        term1 = term1.sel(**{self.LonIndexer: self.eastern_limit}) - term1.sel(
            **{self.LonIndexer: self.western_limit}
        )
        term1 = term1.integrate("rlats")
        term1 = self._handle_nans(term1)
        term1 = (
            term1.integrate(self.VerticalCoordIndexer)
            * self.PressureData.metpy.units
            * self.c1
        )

        # Second term
        term2 = (self.v_ZA * self.geopt_AE) * self.v_ZA["coslats"] / g
        term2 = self._handle_nans(term2)
        term2 = term2.sel(**{self.LatIndexer: self.northern_limit}) - term2.sel(
            **{self.LatIndexer: self.southern_limit}
        )
        term2 = (
            term2.integrate(self.VerticalCoordIndexer)
            * self.PressureData.metpy.units
            * self.c2
        )

        # Third term
        term3 = (
            CalcAreaAverage(
                self.omega_ZE * self.geopt_ZE, self.ylength, xlength=self.xlength
            )
            / g
        )
        term3 = self._handle_nans(term3)
        term3 = term3.isel(**{self.VerticalCoordIndexer: -1}) - term3.isel(
            **{self.VerticalCoordIndexer: 0}
        )

        function = term1 + term2 - term3
        function = self._handle_nans(function)
        Boe = self._convert_units(function, "BOE")

        self.app_logger.debug("Done.")
        return Boe

    def _handle_nans(self, function):
        """
        If there are any, interpolate them and drop any remaining NaN values.
        If there are still NaN values after interpolation, drop the dimensions that contain NaN values.

        Parameters:
            function (np.ndarray): The function to handle NaN values for.

        Returns:
            None
        """
        if np.isnan(function).any():
            function = (
                function.interpolate_na(dim=self.VerticalCoordIndexer)
                * function.metpy.units
            )
            if np.isnan(function).any():
                function = function.dropna(dim=self.VerticalCoordIndexer)
        return function

    def _convert_units(self, function, variable_name):
        """
        Converts the units of a given function to 'J/m^2'.

        Parameters:
            function (type): The function to be converted.
            variable_name (type): The name of the variable associated with the function.

        Raises:
            ValueError: If there is a unit error in the given variable.

        Returns:
            None
        """
        try:
            function = function.metpy.convert_units("W/m^2")
        except ValueError as e:
            error_message = f"Unit error in {variable_name}"
            self.app_logger.exception(error_message)
            raise ValueError(error_message) from e

        return function



import shutil

from lorenzcycletoolkit import (create_arg_parser, initialize_logging,
                                prepare_data, run_lec_analysis,
                                setup_results_directory)


def test_example_args(monkeypatch):

    # Prepare input data
    shutil.copy('inputs/box_limits_Reg1', 'inputs/box_limits')
    shutil.copy('inputs/namelist_ERA5', 'inputs/namelist')

    test_args = ['lorenzcycletoolkit.py', 'samples/testdata_ERA5.nc', '-r', '-f', '-p', '-v']
    monkeypatch.setattr('sys.argv', test_args)

    parser = create_arg_parser()
    args = parser.parse_args()

    assert args.residuals
    assert args.fixed
    assert args.plots
    assert args.verbosity
    assert args.infile == 'samples/testdata_ERA5.nc'

    method = 'fixed'

    # Setup results directory
    results_subdirectory, figures_directory, results_subdirectory_vertical_levels = setup_results_directory(args, method)

    # Initialize logging
    app_logger = initialize_logging(results_subdirectory, args)
    app_logger.info("Starting LEC analysis")
    app_logger.info(f"Command line arguments: {args}")

    # Prepare data
    data = prepare_data(args, 'inputs/namelist', app_logger)

    # Run LEC analysis
    run_lec_analysis(data, args, results_subdirectory, figures_directory, results_subdirectory_vertical_levels, app_logger)



# **************************************************************************** #
#                                                                              #
#                                                         :::      ::::::::    #
#    conversion_terms.py                                :+:      :+:    :+:    #
#                                                     +:+ +:+         +:+      #
#    By: daniloceano <danilo.oceano@gmail.com>      +#+  +:+       +#+         #
#                                                 +#+#+#+#+#+   +#+            #
#    Created: 2022/01/31 20:15:59 by daniloceano       #+#    #+#              #
#    Updated: 2024/07/18 00:20:21 by daniloceano      ###   ########.fr        #
#                                                                              #
# **************************************************************************** #

"""
This script defines the ConversionTerms object. It uses the MetData object as
an input. The built-in functions use the input data to compute the following
energy conversion terms of the Lorenz Energy Cycle.

Created by:
    Danilo Couto de Souza
    Universidade de São Paulo (USP)
    Instituto de Astornomia, Ciências Atmosféricas e Geociências
    São Paulo - Brazil

Contact:
    danilo.oceano@gmail.com
"""

import logging

import numpy as np
from metpy.constants import Rd, Re, g

from ..utils.box_data import BoxData
from ..utils.calc_averages import CalcAreaAverage


class ConversionTerms:
    """
    Class to compute energy conversion terms of the Lorenz Energy Cycle.

    Attributes:
        method (str): The computation method used ('fixed', 'track', or 'choose').
        box_obj (BoxData): The BoxData object containing meteorological data.

    Methods:
        calc_ce: Computes the eedy energy conversion term (CE).
        calc_cz: Computes the zonal energy conversion term (CZ).
        calc_ca: Computes the available potential energy conversion term (CA).
        calc_ck: Computes the kinetic energy conversion term (CK).

    Source for formulas used here:
        Brennan, F. E., & Vincent, D. G. (1980).
        Zonal and Eddy Components of the Synoptic-Scale Energy Budget
        during Intensification of Hurricane Carmen (1974),
        Monthly Weather Review, 108(7), 954-965. Retrieved Jan 25, 2022, from:
        https://journals.ametsoc.org/view/journals/mwre/108/7/1520-0493_1980_108_0954_zaecot_2_0_co_2.xml
    """

    def __init__(self, box_obj: BoxData, method: str, app_logger: logging.Logger):
        """Initialize the ConversionTerms object with a BoxData object and a method."""
        self._initialize_attributes(box_obj, method, app_logger)

    def _initialize_attributes(self, box_obj, method, app_logger):
        """Helper method to initialize attributes from the BoxData object."""
        # Operational attributes
        self.method = method
        self.box_obj = box_obj
        self.results_subdirectory = box_obj.results_subdirectory
        self.results_subdirectory_vertical_levels = box_obj.results_subdirectory_vertical_levels
        self.app_logger = app_logger

        # Initialize spatial and temporal attributes
        self.LonIndexer = box_obj.LonIndexer
        self.LatIndexer = box_obj.LatIndexer
        self.tan_lats = np.tan(box_obj.tair["rlats"])
        self.VerticalCoordIndexer = box_obj.VerticalCoordIndexer
        self.TimeName = box_obj.TimeName

        # Initialize lengths for averaging
        self.xlength = box_obj.xlength
        self.ylength = box_obj.ylength

        # Initialize attributes related to temperature
        self.tair_AE = box_obj.tair_AE
        self.tair_ZE = box_obj.tair_ZE

        # Initialize attributes related to wind components
        self.u_ZA = box_obj.u_ZA
        self.u_ZE = box_obj.u_ZE
        self.v_ZA = box_obj.v_ZA
        self.v_ZE = box_obj.v_ZE

        # Initialize attirbutes related to vertical velocity
        self.omega_ZE = box_obj.omega_ZE
        self.omega_AE = box_obj.omega_AE

        # Initialize the static stability parameter
        self.sigma_AA = box_obj.sigma_AA

        # Initialize the pressure data
        self.PressureData = box_obj.PressureData

    def calc_ca(self):
        """
        Computes conversion between the two available potential energy forms (AZ and AE).

        Note: on the first term of the integral, on Brennan et al. (1980), it is missing
        a 2 in the multiplication Re * self.sigma_AA. This is confirmed by looking to the
        paper by Muench (1965).
        """
        self.app_logger.debug("Calculating CA...")

        # First term of the integral
        DelPhi_tairAE = (self.tair_AE * self.tair_AE["coslats"]).differentiate("rlats")
        term1 = (self.v_ZE * self.tair_ZE * DelPhi_tairAE) / (2 * Re * self.sigma_AA)
        term1 = CalcAreaAverage(term1, self.ylength, xlength=self.xlength)
        self._save_vertical_levels(term1, "Ca_1")

        # Second term of the integral
        DelPres_tairAE = (self.tair_AE).differentiate(
            self.VerticalCoordIndexer
        ) / self.PressureData.metpy.units
        term2 = (self.omega_ZE * self.tair_ZE) * DelPres_tairAE
        term2 = (
            CalcAreaAverage(term2, self.ylength, xlength=self.xlength) / self.sigma_AA
        )
        self._save_vertical_levels(term2, "Ca_2")

        # Process the integral and save the result
        function = -(term1 + term2)
        function = self._handle_nans(function)
        self._save_vertical_levels(function, "Ca")
        Ca = (
            function.integrate(self.VerticalCoordIndexer)
            * self.PressureData.metpy.units
        )
        Ca = self._convert_units(Ca, "Ca")

        self.app_logger.debug("Done.")
        return Ca

    def calc_ce(self):
        """Computes conversion between the two eddy energy forms (AE and KE)."""
        self.app_logger.debug("Calculating CE...")

        # First term of the integral
        term1 = Rd / (self.PressureData * g)
        self._save_vertical_levels(term1, "Ce_1")

        # Second term of the integral
        omega_tair_product = self.omega_ZE * self.tair_ZE
        term2 = CalcAreaAverage(omega_tair_product, self.ylength, xlength=self.xlength)
        self._save_vertical_levels(term2, "Ce_2")

        # Process the integral and save the result
        function = -(term1 * term2)
        function = self._handle_nans(function)
        self._save_vertical_levels(function, "Ce")
        Ce = (
            function.integrate(self.VerticalCoordIndexer)
            * self.PressureData.metpy.units
        )
        Ce = self._convert_units(Ce, "Ce")

        self.app_logger.debug("Done.")
        return Ce

    def calc_cz(self):
        """Computes conversion between the two zonal energy forms (ZE and KE)."""
        self.app_logger.debug("Calculating CZ...")

        # First term of the integral
        term1 = Rd / (self.PressureData * g)
        self._save_vertical_levels(term1, "Cz_1")

        # Second term of the integral
        omega_tair_product = self.omega_AE * self.tair_AE
        term2 = CalcAreaAverage(omega_tair_product, self.ylength)
        self._save_vertical_levels(term2, "Cz_2")

        # Process the integral and save the result
        function = -(term1 * term2)
        function = self._handle_nans(function)
        self._save_vertical_levels(function, "Cz")
        Cz = (
            function.integrate(self.VerticalCoordIndexer)
            * self.PressureData.metpy.units
        )
        Cz = self._convert_units(Cz, "Cz")

        self.app_logger.debug("Done.")
        return Cz

    def calc_ck(self):
        """Computes conversion between the two eddy kinetic energy forms (KE and KZ)."""
        self.app_logger.debug("Calculating CK...")

        # First term of the integral
        DelPhi_uZA_cosphi = (self.u_ZA / self.u_ZA["coslats"]).differentiate("rlats")
        term1 = (self.u_ZE["coslats"] * self.u_ZE * self.v_ZE / Re) * DelPhi_uZA_cosphi
        term1 = CalcAreaAverage(term1, self.ylength, xlength=self.xlength)
        self._save_vertical_levels(term1, "Ck_1")

        # Second term of the integral
        DelPhi_vZA = (self.v_ZA).differentiate("rlats")
        term2 = ((self.v_ZE**2) / Re) * DelPhi_vZA
        term2 = CalcAreaAverage(term2, self.ylength, xlength=self.xlength)
        self._save_vertical_levels(term2, "Ck_2")

        # Third term of the integral
        term3 = (self.tan_lats * (self.u_ZE**2) * self.v_ZA) / Re
        term3 = CalcAreaAverage(term3, self.ylength, xlength=self.xlength)
        self._save_vertical_levels(term3, "Ck_3")

        # Fourth term of the integral
        DelPres_uZAp = (
            self.u_ZA.differentiate(self.VerticalCoordIndexer)
            / self.PressureData.metpy.units
        )
        term4 = self.omega_ZE * self.u_ZE * DelPres_uZAp
        term4 = CalcAreaAverage(term4, self.ylength, xlength=self.xlength)
        self._save_vertical_levels(term4, "Ck_4")

        # Fifth term of the integral
        DelPres_vZAp = (
            self.u_ZA.differentiate(self.VerticalCoordIndexer)
            / self.PressureData.metpy.units
        )
        term5 = self.omega_ZE * self.v_ZE * DelPres_vZAp
        term5 = CalcAreaAverage(term5, self.ylength, xlength=self.xlength)
        self._save_vertical_levels(term5, "Ck_5")

        # Process the integral and save the result
        function = term1 + term2 + term3 + term4 + term5
        function = self._handle_nans(function)
        self._save_vertical_levels(function, "Ck")
        Ck = (
            function.integrate(self.VerticalCoordIndexer)
            * self.PressureData.metpy.units
            / g
        )
        Ck = self._convert_units(Ck, "Ck")

        self.app_logger.debug("Done.")
        return Ck

    def _handle_nans(self, function):
        """
        If there are any, interpolate them and drop any remaining NaN values.
        If there are still NaN values after interpolation, drop the dimensions that contain NaN values.

        Parameters:
            function (np.ndarray): The function to handle NaN values for.

        Returns:
            None
        """
        if np.isnan(function).any():
            function = (
                function.interpolate_na(dim=self.VerticalCoordIndexer)
                * function.metpy.units
            )
            if np.isnan(function).any():
                function = function.dropna(dim=self.VerticalCoordIndexer)
        return function

    def _convert_units(self, function, variable_name):
        """
        Converts the units of a given function to 'J/m^2'.

        Parameters:
            function (type): The function to be converted.
            variable_name (type): The name of the variable associated with the function.

        Raises:
            ValueError: If there is a unit error in the given variable.

        Returns:
            None
        """
        try:
            function = function.metpy.convert_units("W/m^2")
        except ValueError as e:
            raise ValueError(f"Unit error in {variable_name}") from e
        return function

    def _save_vertical_levels(self, function, variable_name):
        """Save computed energy data to a CSV file."""
        df = function.to_dataframe(name=variable_name)
        df.reset_index(inplace=True)

        if self.method == "fixed":
            if self.TimeName not in function.dims:
                df = df.T
            else:
                df = df.pivot(index=self.TimeName, columns=self.VerticalCoordIndexer)

        else:
            df.set_index(self.TimeName, inplace=True)
            df.index = df.index.strftime("%Y-%m-%d %H:%M:%S")
            df = df.pivot(columns=self.VerticalCoordIndexer, values=variable_name)
            df.columns.name = None

        df.to_csv(
            f"{self.results_subdirectory_vertical_levels}/{variable_name}_{self.VerticalCoordIndexer}.csv",
            mode="a",
            header=None,
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Apr 25 17:12:34 2022

This script reads CSV files with energy and conversion terms from the Lorenz
Energy Cycle, for each vertical level and time step, and plots them.

Created by:
    Danilo Couto de Souza
    Universidade de São Paulo (USP)
    Instituto de Astornomia, Ciências Atmosféricas e Geociências
    São Paulo - Brazil

Contact:
    danilo.oceano@gmail.com
"""

import os

import matplotlib.colors as colors
import matplotlib.dates as mdates
import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import numpy as np
import pandas as pd
import src.plots.utils as utils
from src.plots.utils import get_data_vertical_levels, read_results


def _plotter(dict_vertical, figures_subdirectory, app_logger=False):

    times = pd.to_datetime(dict_vertical["Az"].index)
    levs = dict_vertical["Az"].columns

    term_details = utils.TERM_DETAILS
    all_terms = {
        key: [term for term in details["terms"] if term in dict_vertical]
        for key, details in term_details.items()
        if key in ["energy", "conversion", "generation_dissipation"]
    }

    for term_type, terms in all_terms.items():
        if not terms:
            continue  # Skip if no terms to plot
        fig, gs = create_figure_for_term(term_type)

        for i, term in enumerate(terms):
            ax = fig.add_subplot(gs[i])
            plot_contour(times, levs, dict_vertical[term], ax, term)
            finalize_plot(ax, term)

        outfile = os.path.join(figures_subdirectory, f"hovmoller_{term_type}.png")
        plt.savefig(outfile, bbox_inches="tight")
        (
            app_logger.info(f"Figure saved to {outfile}")
            if app_logger
            else print(f"Figure saved to {outfile}")
        )


def create_figure_for_term(term_type):
    if term_type == "energy":
        fig = plt.figure(figsize=(10, 10))
        gs = gridspec.GridSpec(nrows=2, ncols=2, wspace=0.5, hspace=0.3, top=0.95)
    elif term_type == "conversion":
        fig = plt.figure(figsize=(10, 10))
        gs = gridspec.GridSpec(nrows=2, ncols=2, wspace=0.75, hspace=0.3, top=0.95)
    elif term_type == "generation_dissipation":
        fig = plt.figure(figsize=(10, 5))
        gs = gridspec.GridSpec(nrows=1, ncols=2, right=1.2, wspace=0.4)

    return fig, gs


def plot_contour(times, levs, data, ax, term):
    imin = data.min(numeric_only=True).min()
    imax = data.max(numeric_only=True).max()
    absmax = np.amax([np.abs(imin), imax])
    cmap, norm, title = get_cmap_norm_title_for_term(term, imin, imax, absmax)

    cf = ax.contourf(
        times, levs, data[levs].transpose(), cmap=cmap, extend="neither", norm=norm
    )
    ax.contour(times, levs, data[levs].transpose(), colors="k", extend="neither")

    def fmt(x, pos):
        a, b = "{:.2e}".format(x).split("e")
        b = int(b)
        return r"${} \times 10^{{{}}}$".format(a, b)

    if (
        term in utils.TERM_DETAILS["conversion"]["terms"]
        or term in utils.TERM_DETAILS["generation_dissipation"]["terms"]
    ):
        cbar = plt.colorbar(cf, ax=ax, format=ticker.FuncFormatter(fmt))
    else:
        cbar = plt.colorbar(cf, ax=ax)
    cbar.set_label(title, rotation=270, labelpad=20, fontsize=14)
    cbar.ax.tick_params(labelsize=12)

    ax.invert_yaxis()


def get_cmap_norm_title_for_term(term, imin, imax, absmax):
    if term in utils.TERM_DETAILS["energy"]["terms"]:
        cmap = "cmo.amp"
        norm = colors.Normalize(vmin=imin, vmax=imax)
        title = f"{utils.TERM_DETAILS['energy']['label']} ({utils.TERM_DETAILS['energy']['unit']})"
    elif term in utils.TERM_DETAILS["conversion"]["terms"]:
        cmap = "cmo.tarn"
        norm = colors.TwoSlopeNorm(vcenter=0, vmin=-absmax, vmax=absmax)
        title = f'{utils.TERM_DETAILS["conversion"]["label"]} ({utils.TERM_DETAILS["conversion"]["unit"]})'
    elif term in utils.TERM_DETAILS["generation_dissipation"]["terms"]:
        cmap = "cmo.tarn"
        norm = colors.TwoSlopeNorm(vcenter=0, vmin=-absmax, vmax=absmax)
        title = (
            f'{utils.TERM_DETAILS["generation_dissipation"]["label"]} '
            f'({utils.TERM_DETAILS["generation_dissipation"]["unit"]})'
        )
    return cmap, norm, title


def finalize_plot(ax, term):
    locator = mdates.AutoDateLocator(minticks=5, maxticks=12)
    formatter = mdates.ConciseDateFormatter(locator)
    ax.xaxis.set_major_locator(locator)
    ax.xaxis.set_major_formatter(formatter)
    ax.tick_params(axis="x", labelrotation=20)
    ax.xaxis.set_tick_params(labelsize=16)
    ax.yaxis.set_tick_params(labelsize=16)
    ax.set_title(term, fontsize=20)


def plot_hovmoller(results_file, figures_directory, app_logger=False):

    read_results(results_file, app_logger)

    results_directory = os.path.dirname(results_file)
    dict_vertical = get_data_vertical_levels(results_directory)

    figures_subdirectory = os.path.join(figures_directory, "boxplots")
    os.makedirs(figures_subdirectory, exist_ok=True)

    figures_directory = os.path.join(results_directory, "Figures")
    figures_subdirectory = os.path.join(figures_directory, "hovmollers")
    os.makedirs(figures_subdirectory, exist_ok=True)

    (
        app_logger.info("Creating hovmoller diagrams")
        if app_logger
        else print("Creating hovmoller diagrams")
    )
    _plotter(dict_vertical, figures_subdirectory, app_logger)
    (
        app_logger.info("Hovmoller diagrams created")
        if app_logger
        else print("Hovmoller diagrams created")
    )


if __name__ == "__main__":

    # Test for Reg1-Representative_fixed
    results_file = "samples/Reg1-Representative_NCEP-R2_fixed/Reg1-Representative_NCEP-R2_fixed_results.csv"
    figures_directory = "samples/Reg1-Representative_fixed/Figures/"
    plot_hovmoller(results_file, figures_directory)

    # Test for Catarina_NCEP-R2_fixed
    results_file = "samples/Catarina_NCEP-R2_fixed/Catarina_NCEP-R2_fixed_results.csv"
    figures_directory = "samples/Catarina_NCEP-R2_fixed/Figures/"
    plot_hovmoller(results_file, figures_directory)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jan  6 12:33:00 2023

Created by:
    Danilo Couto de Souza
    Universidade de São Paulo (USP)
    Instituto de Astornomia, Ciências Atmosféricas e Geociências
    São Paulo - Brazil

Contact:
    danilo.oceano@gmail.com
"""

import logging
import os
import time

import cartopy.crs as ccrs
import cartopy.feature as cfeature
import cmocean.cm as cmo
import matplotlib.colors as colors
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from cartopy.feature import BORDERS, COASTLINE, NaturalEarthFeature
from metpy.calc import vorticity
from shapely.geometry.polygon import Polygon

nclicks = 2

# define all CRS
crs_longlat = ccrs.PlateCarree()

# Transformation function


def coordXform(orig_crs, target_crs, x, y):
    return target_crs.transform_points(orig_crs, x, y)


def tellme(s):
    plt.title(s, fontsize=16)
    plt.draw()


def draw_box(ax, limits, crs):
    max_lon, min_lon = limits["max_lon"], limits["min_lon"]
    max_lat, min_lat = limits["max_lat"], limits["min_lat"]
    pgon = Polygon(
        (
            (min_lon, min_lat),
            (min_lon, max_lat),
            (max_lon, max_lat),
            (max_lon, min_lat),
            (min_lon, min_lat),
        )
    )
    ax.add_geometries(
        [pgon], crs=crs, facecolor="None", edgecolor="k", linewidth=3, alpha=1, zorder=3
    )


def plot_zeta(ax, zeta, lat, lon, hgt=None):
    if np.abs(zeta.min()) < np.abs(zeta.max()):
        norm = colors.TwoSlopeNorm(vmin=-zeta.max(), vcenter=0, vmax=zeta.max())
    else:
        norm = colors.TwoSlopeNorm(vmin=zeta.min(), vcenter=0, vmax=-zeta.min())
    cmap = cmo.balance
    # plot contours
    cf1 = ax.contourf(
        lon, lat, zeta, cmap=cmap, norm=norm, levels=51, transform=crs_longlat
    )
    plt.colorbar(cf1, pad=0.12, orientation="vertical", shrink=0.5)
    if hgt is not None:
        cs = ax.contour(
            lon,
            lat,
            hgt,
            levels=11,
            colors="#344e41",
            linestyles="dashed",
            linewidths=1.3,
            transform=crs_longlat,
        )
        ax.clabel(cs, cs.levels, inline=True, fontsize=10)


def map_decorators(ax):
    ax.coastlines()
    gl = ax.gridlines(
        draw_labels=True,
        zorder=2,
        linestyle="dashed",
        alpha=0.7,
        linewidth=0.5,
        color="#383838",
    )
    gl.xlabel_style = {"size": 14, "color": "#383838"}
    gl.ylabel_style = {"size": 14, "color": "#383838"}
    gl.top_labels = None
    gl.right_labels = None


def plot_min_zeta(ax, zeta, lat, lon, limits):
    max_lon, min_lon = limits["max_lon"], limits["min_lon"]
    max_lat, min_lat = limits["max_lat"], limits["min_lat"]

    # Plot mininum zeta point whithin box
    izeta = zeta.sel(
        {lon.dims[0]: slice(min_lon, max_lon), lat.dims[0]: slice(min_lat, max_lat)}
    )
    if min_lat < 0:
        min_max_zeta = izeta.min()
    else:
        min_max_zeta = izeta.max()

    min_max_zeta_loc = izeta.where(izeta == min_max_zeta, drop=True).squeeze()

    # sometimes there are multiple minimuns
    if min_max_zeta_loc.shape:
        if len(min_max_zeta_loc.shape) > 1:
            for points in min_max_zeta_loc:
                for point in points:
                    ax.scatter(
                        point[lon.dims[0]],
                        point[lat.dims[0]],
                        marker="o",
                        facecolors="none",
                        linewidth=3,
                        edgecolor="k",
                        s=200,
                    )
        else:
            for point in min_max_zeta_loc:
                ax.scatter(
                    point[lon.dims[0]],
                    point[lat.dims[0]],
                    marker="o",
                    facecolors="none",
                    linewidth=3,
                    edgecolor="k",
                    s=200,
                )
    else:
        ax.scatter(
            min_max_zeta_loc[lon.dims[0]],
            min_max_zeta_loc[lat.dims[0]],
            marker="o",
            facecolors="none",
            linewidth=3,
            edgecolor="k",
            s=200,
        )


def initial_domain(zeta, lat, lon):
    plt.close("all")
    fig = plt.figure(figsize=(10, 8))
    ax = plt.axes(projection=crs_longlat)
    fig.add_axes(ax)
    ax.set_global()
    plot_zeta(ax, zeta.compute(), lat, lon)
    map_decorators(ax)
    plt.subplots_adjust(bottom=0, top=1.2)

    while True:
        pts = []
        while len(pts) < nclicks:
            tellme("Select an initial spatial domain")
            pts = np.asarray(
                plt.ginput(nclicks, timeout=15, mouse_stop="MouseButton.MIDDLE")
            )
            if len(pts) < nclicks:
                tellme("Too few points, starting over")
                time.sleep(1)  # Wait a second

        xmin, xmax = min([pts[0, 0], pts[1, 0]]), max([pts[0, 0], pts[1, 0]])
        ymin, ymax = min([pts[0, 1], pts[1, 1]]), max([pts[0, 1], pts[1, 1]])
        xs, ys = np.array((xmin, xmax)), np.array((ymin, ymax))
        lls = coordXform(crs_longlat, crs_longlat, xs, ys)
        min_lon, min_lat = lls[0, 0], lls[0, 1]
        max_lon, max_lat = lls[1, 0], lls[1, 1]

        limits = {
            "min_lon": min_lon,
            "max_lon": max_lon,
            "min_lat": min_lat,
            "max_lat": max_lat,
        }
        draw_box(ax, limits, crs_longlat)

        tellme("Happy? Key press any keyboard key for yes, mouse click for no")
        if plt.waitforbuttonpress():
            break

    return limits


def draw_box_map(u, v, zeta, hgt, lat, lon, timestr):
    plt.close("all")
    fig = plt.figure(figsize=(10, 8))
    ax = plt.axes(projection=crs_longlat)
    fig.add_axes(ax)

    plot_zeta(ax, zeta, lat, lon, hgt)
    ax.streamplot(
        lon.values,
        lat.values,
        u.values,
        v.values,
        color="#2A1D21",
        transform=crs_longlat,
    )
    map_decorators(ax)

    while True:
        pts = []
        while len(pts) < nclicks:
            time_formated = timestr.strftime("%Y-%m-%d %HZ")
            tellme(f"Select box corners \nModel time step: {time_formated}")
            pts = np.asarray(
                plt.ginput(nclicks, timeout=15, mouse_stop="MouseButton.MIDDLE")
            )
            if len(pts) < nclicks:
                tellme("Too few points, starting over")
                time.sleep(1)  # Wait a second

        xmin, xmax = min([pts[0, 0], pts[1, 0]]), max([pts[0, 0], pts[1, 0]])
        ymin, ymax = min([pts[0, 1], pts[1, 1]]), max([pts[0, 1], pts[1, 1]])
        xs, ys = np.array((xmin, xmax)), np.array((ymin, ymax))
        lls = coordXform(crs_longlat, crs_longlat, xs, ys)
        min_lon, min_lat = lls[0, 0], lls[0, 1]
        max_lon, max_lat = lls[1, 0], lls[1, 1]

        limits = {
            "min_lon": min_lon,
            "max_lon": max_lon,
            "min_lat": min_lat,
            "max_lat": max_lat,
        }
        draw_box(ax, limits, crs_longlat)
        plot_min_zeta(ax, zeta, lat, lon, limits)

        tellme("Happy? Key press any keyboard key for yes, mouse click for no")

        if plt.waitforbuttonpress():
            break

    return limits


def slice_domain(NetCDF_data, args, variable_list_df):
    """
    Slice a NetCDF dataset based on given limits and return the sliced dataset.

    Parameters:
        NetCDF_data (xarray.Dataset): The NetCDF dataset to be sliced.
        args (argparse.Namespace): Command-line arguments.
        variable_list_df (pandas.DataFrame): The DataFrame containing the variable list.

    Returns:
        xarray.Dataset: The sliced NetCDF dataset.
    """
    # Data indexers
    LonIndexer = variable_list_df.loc["Longitude"]["Variable"]
    LatIndexer = variable_list_df.loc["Latitude"]["Variable"]
    TimeIndexer = variable_list_df.loc["Time"]["Variable"]
    LevelIndexer = variable_list_df.loc["Vertical Level"]["Variable"]

    if args.fixed:
        dfbox = pd.read_csv(
            "inputs/box_limits", header=None, delimiter=";", index_col=0
        )
        WesternLimit = float(
            NetCDF_data[LonIndexer].sel(
                {LonIndexer: float(dfbox.loc["min_lon"].values[0])}, method="nearest"
            )
        )
        EasternLimit = float(
            NetCDF_data[LonIndexer].sel(
                {LonIndexer: float(dfbox.loc["max_lon"].values[0])}, method="nearest"
            )
        )
        SouthernLimit = float(
            NetCDF_data[LatIndexer].sel(
                {LatIndexer: float(dfbox.loc["min_lat"].values[0])}, method="nearest"
            )
        )
        NorthernLimit = float(
            NetCDF_data[LatIndexer].sel(
                {LatIndexer: float(dfbox.loc["max_lat"].values[0])}, method="nearest"
            )
        )

    elif args.track:
        dx, dy = (
            NetCDF_data[LonIndexer][1] - NetCDF_data[LonIndexer][0],
            NetCDF_data[LatIndexer][1] - NetCDF_data[LatIndexer][0],
        )

        trackfile = args.trackfile if args.trackfile else "inputs/track"
        track = pd.read_csv(trackfile, parse_dates=[0], delimiter=";", index_col="time")
        if "width" in track.columns:
            max_width, max_length = track["width"].max(), track["length"].max()
        else:
            max_width, max_length = 15, 15

        WesternLimit = track["Lon"].min() - (max_width / 2) - dx
        EasternLimit = track["Lon"].max() + (max_width / 2) + dx
        SouthernLimit = track["Lat"].min() - (max_length / 2) - dy
        NorthernLimit = track["Lat"].max() + (max_length / 2) + dy

    elif args.choose:
        iu_850 = NetCDF_data.isel({TimeIndexer: 0}).sel(
            {LevelIndexer: 8500}, method="nearest"
        )[variable_list_df.loc["Eastward Wind Component"]["Variable"]]
        iv_850 = NetCDF_data.isel({TimeIndexer: 0}).sel(
            {LevelIndexer: 8500}, method="nearest"
        )[variable_list_df.loc["Northward Wind Component"]["Variable"]]
        zeta = vorticity(iu_850, iv_850).metpy.dequantify()

        lat, lon = iu_850[LatIndexer], iu_850[LonIndexer]
        domain_limits = initial_domain(zeta, lat, lon)
        WesternLimit = domain_limits["min_lon"]
        EasternLimit = domain_limits["max_lon"]
        SouthernLimit = domain_limits["min_lat"]
        NorthernLimit = domain_limits["max_lat"]

    NetCDF_data = NetCDF_data.sel(
        **{
            LatIndexer: slice(SouthernLimit, NorthernLimit),
            LonIndexer: slice(WesternLimit, EasternLimit),
        }
    )

    return NetCDF_data


def plot_domain_attributes(data850, limits, position, figures_directory):

    time = limits["datestr"]
    central_lon, central_lat = limits["central_lon"], limits["central_lat"]
    width, length = limits["width"], limits["length"]
    min_lon = central_lon - (width / 2)
    max_lon = central_lon + (width / 2)
    min_lat = central_lat - (length / 2)
    max_lat = central_lat + (length / 2)

    # Create figure
    plt.close("all")
    fig, ax = plt.subplots(
        figsize=(8, 8.5), subplot_kw=dict(projection=ccrs.PlateCarree())
    )

    # Set map extent and features
    ax.set_extent(
        [min_lon - 20, max_lon + 20, max_lat + 20, min_lat - 20], crs=ccrs.PlateCarree()
    )

    # Plot selected domain
    # Create a sample polygon, `pgon`
    pgon = Polygon(
        (
            (min_lon, min_lat),
            (min_lon, max_lat),
            (max_lon, max_lat),
            (max_lon, min_lat),
            (min_lon, min_lat),
        )
    )
    ax.add_geometries(
        [pgon],
        crs=ccrs.PlateCarree(),
        facecolor="None",
        edgecolor="gray",
        linewidth=3,
        alpha=1,
        zorder=3,
    )

    # Plot central point, mininum vorticity, minimum hgt and maximum wind
    ax.scatter(central_lon, central_lat, marker="o", c="#31332e", s=100, zorder=4)
    ax.scatter(
        position["min_max_zeta_850"]["longitude"],
        position["min_max_zeta_850"]["latitude"],
        marker="s",
        c="#31332e",
        s=100,
        zorder=4,
        label="min/max zeta",
    )
    ax.scatter(
        position["min_hgt_850"]["longitude"],
        position["min_hgt_850"]["latitude"],
        marker="x",
        c="#31332e",
        s=100,
        zorder=4,
        label="min hgt",
    )
    ax.scatter(
        position["max_wind_850"]["longitude"],
        position["max_wind_850"]["latitude"],
        marker="^",
        c="#31332e",
        s=100,
        zorder=4,
        label="max wind",
    )

    plt.legend(loc="upper left", frameon=True, fontsize=14, bbox_to_anchor=(1.1, 1.2))

    # Add gridlines
    gl = ax.gridlines(draw_labels=True, zorder=2)
    gl.xlabel_style = {"size": 16}
    gl.ylabel_style = {"size": 16}

    # Add title
    plt.title("Box defined for computations\n", fontsize=22)

    plot_zeta(
        ax, data850["izeta_850"], data850["lat"], data850["lon"], data850["izeta_850"]
    )
    ax.add_feature(COASTLINE, edgecolor="#283618", linewidth=1)
    ax.add_feature(BORDERS, edgecolor="#283618", linewidth=1)
    _ = ax.add_feature(
        cfeature.NaturalEarthFeature(
            "physical", "land", "50m", edgecolor="face", facecolor="None"
        )
    )

    states = NaturalEarthFeature(
        category="cultural",
        scale="50m",
        facecolor="none",
        name="admin_1_states_provinces_lines",
    )
    _ = ax.add_feature(states, edgecolor="#283618", linewidth=1)

    cities = NaturalEarthFeature(
        category="cultural", scale="50m", facecolor="none", name="populated_places"
    )
    _ = ax.add_feature(cities, edgecolor="#283618", linewidth=1)

    # Save figure
    boxes_directory = os.path.join(figures_directory, "boxes")
    os.makedirs(boxes_directory, exist_ok=True)
    filename = os.path.join(boxes_directory, f"box_{time}.png")
    plt.savefig(filename)
    logging.debug(f"\nCreated figure with box defined for computations at {filename}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Script for computing zonal and area averages in a sphere.

Source:
    Brennan, F. E., & Vincent, D. G. (1980).
    Zonal and Eddy Components of the Synoptic-Scale Energy Budget
    during Intensification of Hurricane Carmen (1974),
    Monthly Weather Review, 108(7), 954-965. Retrieved Jan 25, 2022, from:
    https://journals.ametsoc.org/view/journals/mwre/108/7/1520-0493_1980_108_0954_zaecot_2_0_co_2.xml

Created by Danilo Couto de Souza
Universidade de São Paulo (USP)
Instituto de Astornomia, Ciências Atmosféricas e Geociências
São Paulo - Brazil

danilo.oceano@gmail.com

"""

import numpy as np


def CalcZonalAverage(VariableData, xlength):
    """
    Computates variable zonal average of some variable, for all z
    levels and time steps.

    Parameters
    ----------
    VariableData: xarray.Dataset
        arrays containing data to be integrated. Requires dimension rlons
        (longitude in radians)
    xlength: float
        Length (in radians), of the data as eastern limit minus western limit.

    Returns
    -------
    ZA: xarray.Dataset
        Arrays of zonal avreages for all longitudes from the passed Dataset
    """
    return VariableData.integrate("rlons") / xlength


def CalcAreaAverage(VariableData, ylength, xlength=False):
    """
    Computates the Area Average of a function.

    The default is to computate the zonal average and then a meridional average.
    If the xlength is provided, it will firstly compute the zonal average and
    then, the area average.

    Parameters
    ----------
    VariableData: xarray.Dataset
        arrays containing data to be integrated
    ylength: float
        Length (in radians), of the data as northern limit minus southern limit.
    xlength: float (optional)
        Length (in radians), of the data as eastern limit minus western limit.
        If passed, it will first compute zonal averages

    Returns
    -------
    AA: xarray.Dataset
        Arrays of area avreages for all latitudes and longitudes from
        the passed Dataset
    """
    # Compute zonal average if requested
    if xlength:
        ZA = CalcZonalAverage(VariableData, xlength)
    else:
        ZA = VariableData
    ylength = np.sin(VariableData["rlats"][-1]) - np.sin(VariableData["rlats"][0])
    return ((ZA * ZA["coslats"]).integrate("rlats") / ylength).drop_vars(
        "coslats", errors="ignore"
    )



# **************************************************************************** #
#                                                                              #
#                                                         :::      ::::::::    #
#    lec_fixed_framework.py                             :+:      :+:    :+:    #
#                                                     +:+ +:+         +:+      #
#    By: daniloceano <danilo.oceano@gmail.com>      +#+  +:+       +#+         #
#                                                 +#+#+#+#+#+   +#+            #
#    Created: 2023/12/19 17:32:59 by daniloceano       #+#    #+#              #
#    Updated: 2024/07/18 00:20:44 by daniloceano      ###   ########.fr        #
#                                                                              #
# **************************************************************************** #

import argparse
import logging
import os
from pathlib import Path

import pandas as pd
import xarray as xr

from ..analysis.boundary_terms import BoundaryTerms
from ..analysis.conversion_terms import ConversionTerms
from ..analysis.energy_contents import EnergyContents
from ..analysis.generation_and_dissipation_terms import \
    GenerationDissipationTerms
from ..utils.box_data import BoxData
from ..utils.calc_budget_and_residual import calc_budget_diff, calc_residuals


def lec_fixed(
    data: xr.Dataset,
    variable_list_df: pd.DataFrame,
    results_subdirectory: str,
    results_subdirectory_vertical_levels: str,
    app_logger: logging.Logger,
    args: argparse.Namespace,
):
    """
    Computes the Lorenz Energy Cycle (LEC) using a fixed framework.

    Args:
        data (xr.Dataset): Dataset containing the atmospheric data for LEC computation.
        variable_list_df (pd.DataFrame): DataFrame with variable mappings used in the LEC analysis.
        results_subdirectory (str): Directory path to save the results.
        results_subdirectory_vertical_levels (str): Directory path to save the vertical level results.
        args (argparse.Namespace): Arguments provided to the script, including configurations
                                   for the LEC computation.

    Raises:
        ValueError: If the bounding box limits are invalid.
        Exception: General exceptions for processing errors.

    Note:
        This function computes various aspects of the LEC and saves the results in CSV format.
        It also triggers plotting scripts if specified in the arguments.
    """
    logging.info("--- Computing energetics using fixed framework ---")

    box_limits_file = args.box_limits
    dfbox = pd.read_csv(box_limits_file, header=None, delimiter=";", index_col=0)
    min_lon, max_lon = dfbox.loc["min_lon"].iloc[0], dfbox.loc["max_lon"].iloc[0]
    min_lat, max_lat = dfbox.loc["min_lat"].iloc[0], dfbox.loc["max_lat"].iloc[0]

    if min_lon > max_lon:
        error_message = f"Error in box_limits: min_lon ({min_lon}) is greater than max_lon ({max_lon})"
        app_logger.error(error_message)
        raise ValueError(error_message)

    if min_lat > max_lat:
        error_message = f"Error in box_limits: min_lat ({min_lat}) is greater than max_lat ({max_lat})"
        app_logger.error(error_message)
        raise ValueError(error_message)

    app_logger.debug("Loading data into memory..")
    data = data.compute()
    app_logger.debug("Data loaded into memory.")

    _, _, TimeName, VerticalCoordIndexer = (
        variable_list_df.loc["Longitude"]["Variable"],
        variable_list_df.loc["Latitude"]["Variable"],
        variable_list_df.loc["Time"]["Variable"],
        variable_list_df.loc["Vertical Level"]["Variable"],
    )

    pres = data[VerticalCoordIndexer] * data[VerticalCoordIndexer].metpy.units
    app_logger.info(
        f"Bounding box: min_lon={min_lon}, max_lon={max_lon}, min_lat={min_lat}, max_lat={max_lat}"
    )

    for term in [
        "Az",
        "Ae",
        "Kz",
        "Ke",
        "Ge",
        "Gz",
        "Cz",
        "Cz_1",
        "Cz_2",
        "Ca",
        "Ca_1",
        "Ca_2",
        "Ce",
        "Ce_1",
        "Ce_2",
        "Ck",
        "Ck_1",
        "Ck_2",
        "Ck_3",
        "Ck_4",
        "Ck_5",
    ]:
        columns = [TimeName] + [float(i) / 100 for i in pres.values]
        output_path = Path(results_subdirectory_vertical_levels, f"{term}_{VerticalCoordIndexer}.csv")
        pd.DataFrame(columns=columns).to_csv(output_path, index=None)

    try:
        box_obj = BoxData(
            data,
            variable_list_df,
            min_lon,
            max_lon,
            min_lat,
            max_lat,
            args,
            results_subdirectory,
            results_subdirectory_vertical_levels,
        )
    except Exception:
        app_logger.exception("An exception occurred while creating BoxData object")
        raise

    try:
        ec_obj = EnergyContents(box_obj, "fixed", app_logger)
        energy_list = [
            ec_obj.calc_az(),
            ec_obj.calc_ae(),
            ec_obj.calc_kz(),
            ec_obj.calc_ke(),
        ]
    except Exception:
        app_logger.exception(
            "An exception occurred while creating EnergyContents object"
        )
        raise
    app_logger.info("Computed energy contents")

    try:
        ct_obj = ConversionTerms(box_obj, "fixed", app_logger)
        conversion_list = [
            ct_obj.calc_cz(),
            ct_obj.calc_ca(),
            ct_obj.calc_ck(),
            ct_obj.calc_ce(),
        ]
    except Exception:
        app_logger.exception(
            "An exception occurred while creating ConversionTerms object"
        )
        raise
    app_logger.info("Computed conversion terms")

    try:
        bt_obj = BoundaryTerms(box_obj, "fixed", app_logger)
        boundary_list = [
            bt_obj.calc_baz(),
            bt_obj.calc_bae(),
            bt_obj.calc_bkz(),
            bt_obj.calc_bke(),
            bt_obj.calc_boz(),
            bt_obj.calc_boe(),
        ]
    except Exception:
        app_logger.exception(
            "An exception occurred while creating BoundaryTerms object"
        )
        raise
    app_logger.info("Computed boundary terms")

    try:
        gdt_obj = GenerationDissipationTerms(box_obj, "fixed", app_logger)
        gen_diss_list = (
            [gdt_obj.calc_gz(), gdt_obj.calc_ge()]
            if args.residuals
            else [
                gdt_obj.calc_gz(),
                gdt_obj.calc_ge(),
                gdt_obj.calc_dz(),
                gdt_obj.calc_de(),
            ]
        )
    except Exception:
        app_logger.exception(
            "An exception occurred while creating GenerationDissipationTerms object"
        )
        raise
    app_logger.info("Computed generation dissipation terms")

    dates = data[TimeName].values
    df = pd.DataFrame(index=dates.astype("datetime64"))
    for i, col in enumerate(["Az", "Ae", "Kz", "Ke"]):
        df[col] = energy_list[i]
    for i, col in enumerate(["Cz", "Ca", "Ck", "Ce"]):
        df[col] = conversion_list[i]
    for i, col in enumerate(
        ["BAz", "BAe", "BKz", "BKe", "Gz", "Ge", "Dz", "De"][: len(gen_diss_list) + 4]
    ):
        df[col] = boundary_list[i] if i < 4 else gen_diss_list[i - 4]

    df = calc_budget_diff(df, dates, app_logger)
    df = calc_residuals(df, app_logger)
    app_logger.info("Computed budget and residuals")

    if args.outname:
        results_filename = args.outname
    else:
        infile_name = os.path.basename(args.infile).split(".nc")[0]
        results_filename = "".join(f"{infile_name}_fixed_results")
    results_file = Path(results_subdirectory, f"{results_filename}.csv")
    df.to_csv(results_file)
    app_logger.info(f"Results saved to {results_file}")

    if args.plots:
        from ..plots.map_box_limits import plot_box_limits
        from ..plots.plot_boxplot import boxplot_terms
        from ..plots.plot_hovmoller import plot_hovmoller
        from ..plots.plot_LEC import plot_lorenzcycletoolkit
        from ..plots.timeseries_terms import plot_timeseries

        app_logger.info("Generating plots..")
        figures_directory = os.path.join(results_subdirectory, "Figures")
        plot_timeseries(results_file, figures_directory, app_logger)
        plot_box_limits(box_limits_file, figures_directory, app_logger)
        boxplot_terms(results_file, results_subdirectory, figures_directory, app_logger)
        plot_hovmoller(results_file, figures_directory, app_logger)
        plot_lorenzcycletoolkit(results_file, figures_directory, app_logger=app_logger)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon May 30 14:22:40 2022

This script reads CSV files with energy and conversion terms from the Lorenz
Energy Cycle and make boxplots.


Created by:
    Danilo Couto de Souza
    Universidade de São Paulo (USP)
    Instituto de Astornomia, Ciências Atmosféricas e Geociências
    São Paulo - Brazil

Contact:
    danilo.oceano@gmail.com
"""

import os
from datetime import datetime

import matplotlib.dates as mdates
import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt

import src.plots.utils as utils
from src.plots.utils import get_data_vertical_levels, read_results


def boxplot_time(dict_vertical, figures_subdirectory, app_logger):
    """
    Generates a boxplot for each term in the given dictionary and saves the resulting figures in the specified subdirectory.

    Parameters:
    - dict_vertical: A dictionary containing vertical data for different terms.
    - figures_subdirectory: The subdirectory where the generated figures will be saved.
    - app_logger: (Optional) The application logger to log information about the saved figures.

    Returns:
    None
    """
    term_details = utils.TERM_DETAILS
    all_terms = {
        key: [term for term in details["terms"] if term in dict_vertical]
        for key, details in term_details.items()
        if key in ["energy", "conversion", "generation_dissipation"]
    }

    dummy_data = dict_vertical[next(iter(all_terms["energy"]), None)]  # Safe access
    t_id = dummy_data.columns[0]  # Time indexer
    times = dummy_data[t_id].index

    for term_type, terms in all_terms.items():
        if not terms:
            continue  # Skip if no terms to plot

        plt.close("all")
        fig_layout = (10, 10) if len(terms) == 4 else (10, 5)
        fig = plt.figure(figsize=fig_layout)
        n_rows, n_cols = (2, 2) if len(terms) == 4 else (1, 2)

        gs = gridspec.GridSpec(nrows=n_rows, ncols=n_cols, hspace=0.25, wspace=0.3)

        for i, term in enumerate(terms):
            ax = fig.add_subplot(gs[i])
            if i % n_cols == 0:
                label = f'{term_details[term_type]["label"]} ({term_details[term_type]["unit"]})'
                ax.set_ylabel(label, fontsize=18)
            for t in times:
                data = dict_vertical[term].loc[t].values
                time_step = mdates.date2num(
                    datetime.fromisoformat(t.strftime("%Y-%m-%d %H:%M"))
                )
                bplot = ax.boxplot(data, positions=[time_step], patch_artist=True)
                bplot["boxes"][-1].set_facecolor(utils.COLORS[i % len(utils.COLORS)])
                bplot["boxes"][-1].set_alpha(0.85)
                for median in bplot["medians"]:
                    median.set_color("k")

            locator = mdates.AutoDateLocator(minticks=5, maxticks=12)
            formatter = mdates.ConciseDateFormatter(locator)
            ax.xaxis.set_major_locator(locator)
            ax.xaxis.set_major_formatter(formatter)
            ax.tick_params(axis="x", labelrotation=20)
            ax.xaxis.set_tick_params(labelsize=16)
            ax.yaxis.set_tick_params(labelsize=16)
            ax.set_title(term, fontsize=20)

        file_name = f"{figures_subdirectory}/vertical_timeseries_{term_type}.png"
        plt.savefig(file_name, bbox_inches="tight")
        (
            app_logger.info(f"Figure saved in directory: {figures_subdirectory}")
            if app_logger
            else print(f"Figure saved in directory: {figures_subdirectory}")
        )
        plt.close("all")


def boxplot_vertical(dict_vertical, figures_subdirectory, app_logger):
    # Define the labels lists based on utils.TERM_DETAILS
    term_details = utils.TERM_DETAILS
    all_terms = {
        key: [term for term in details["terms"] if term in dict_vertical]
        for key, details in term_details.items()
        if key in ["energy", "conversion", "generation_dissipation"]
    }

    linecolors = utils.COLORS

    for term_type, terms in all_terms.items():
        if not terms:
            continue  # Skip if no terms to plot

        plt.close("all")
        fig_layout = (10, 10) if len(terms) == 4 else (10, 5)
        fig = plt.figure(figsize=fig_layout)
        n_rows, n_cols = (2, 2) if len(terms) == 4 else (1, 2)

        wspace = 0.3 if term_type != "generation_dissipation" else 0.4

        gs = gridspec.GridSpec(nrows=n_rows, ncols=n_cols, hspace=0.15, wspace=wspace)

        label = (
            f'{term_details[term_type]["label"]} ({term_details[term_type]["unit"]})'
        )
        for i, term in enumerate(terms):
            ax = fig.add_subplot(gs[i])

            if i % 2 == 0:
                ax.set_ylabel(label, fontsize=18)

            # Assuming second column onward are levels
            levs = dict_vertical[term].columns[1:]
            for j, lev in enumerate(levs):
                bplot = ax.boxplot(
                    dict_vertical[term][lev].values,
                    positions=[j / 3],
                    tick_labels=[lev],
                    patch_artist=True,
                )
                bplot["boxes"][-1].set_facecolor(linecolors[i % len(linecolors)])
                bplot["boxes"][-1].set_alpha(0.85)
                for median in bplot["medians"]:
                    median.set_color("k")

            ax.tick_params(axis="x", labelrotation=60)
            if i < len(terms) - 2:  # Assuming last row should have x labels
                ax.axes.xaxis.set_ticklabels([])
            ax.xaxis.set_tick_params(labelsize=16)
            ax.yaxis.set_tick_params(labelsize=16)
            ax.set_title(term, fontsize=20)

        outfile = f"{figures_subdirectory}/boxplot_vertical_levels_{term_type}.png"
        plt.savefig(outfile, bbox_inches="tight")
        (
            app_logger.info(f"Created {outfile}")
            if app_logger
            else print(f"Created {outfile}")
        )
        plt.close("all")


def plot_boxplot_terms(df, figures_subdirectory, app_logger):
    """
    Generates a boxplot for each term in the term_list from the given DataFrame and saves the resulting figure.

    Parameters:
    - df: The DataFrame containing the data to be plotted.
    - term_list: List of terms to plot.
    - figures_subdirectory: The directory where the figure will be saved.
    - app_logger: Application logger for logging messages.
    """
    # Define the labels lists based on utils.TERM_DETAILS
    term_details = utils.TERM_DETAILS
    all_terms = {
        key: [term for term in details["terms"] if term in df.columns]
        for key, details in term_details.items()
        if key in term_details.keys()
    }

    for term_type, term_list in all_terms.items():
        if not term_list:
            continue  # Skip if no terms to plot

        # Determine the figure size based on the number of terms
        fig_layout = (8, 8) if len(term_list) == 4 else (6, 8)
        plt.figure(figsize=fig_layout)

        for i, term in enumerate(term_list):
            bplot = plt.boxplot(
                df[term],
                positions=[i],
                vert=True,
                patch_artist=True,
                notch=True,
                tick_labels=[term],
            )
            box_color = utils.COLORS[i % len(utils.COLORS)]
            for box in bplot["boxes"]:
                box.set_facecolor(box_color)
                box.set_alpha(0.85)
            for median in bplot["medians"]:
                median.set_color("k")

            # Horizontal line for 0 if not an energy term
            if term not in utils.TERM_DETAILS["energy"]["terms"]:
                plt.axhline(
                    y=0, color="k", linestyle="-", linewidth=1, zorder=1, alpha=0.8
                )

            label = utils.TERM_DETAILS[term_type]["label"]
            plt.ylabel(
                f'{label} ({utils.TERM_DETAILS[term_type]["unit"]})', fontsize=14
            )
            plt.tick_params(axis="x", labelsize=12)
            plt.tick_params(axis="y", labelsize=12)

        # Set x-ticks rotation if budget difference terms are included
        if any(
            term in utils.TERM_DETAILS["budget_diff"]["terms"] for term in term_list
        ):
            plt.xticks(rotation=25)

        plt.tight_layout()

        # Saving figure
        fname = f"boxplot_terms_{term_type}"
        plt.savefig(f"{figures_subdirectory}/{fname}.png")
        (
            app_logger.info(f"{fname}.png created")
            if app_logger
            else print(f"{fname}.png created")
        )
        plt.close("all")


def boxplot_terms(results_file, results_subdirectory, figures_directory, app_logger=False):
    (
        app_logger.info("Plotting boxplots...")
        if app_logger
        else print("Plotting boxplots...")
    )

    df = read_results(results_file, app_logger)

    dict_vertical = get_data_vertical_levels(results_subdirectory)

    figures_subdirectory = os.path.join(figures_directory, "boxplots")
    os.makedirs(figures_subdirectory, exist_ok=True)

    (
        app_logger.info("Creating boxplot for each time")
        if app_logger
        else print("Creating boxplot for each time")
    )
    boxplot_time(dict_vertical, figures_subdirectory, app_logger)

    (
        app_logger.info("Creating boxplot for each vertical level")
        if app_logger
        else print("Creating boxplot for each vertical level")
    )
    boxplot_vertical(dict_vertical, figures_subdirectory, app_logger)

    (
        app_logger.info("Creating boxplot for each term")
        if app_logger
        else print("Creating boxplot for each term")
    )
    plot_boxplot_terms(df, figures_subdirectory, app_logger)


if __name__ == "__main__":
    results_file = "samples/sample_results.csv"
    results_directory = "samples/"
    figures_directory = "samples/Figures/"
    boxplot_terms(results_file, results_directory, figures_directory, app_logger=False)
